import platform
import json
from unittest import TestCase
from packaging import version

import numpy as np
from hdmf.common import DynamicTable, DynamicTableRegion
from pynwb.file import TimeIntervals, Units, ElectrodeTable, ElectrodeGroup, Device

from nwbinspector import (
    InspectorMessage,
    Importance,
    check_empty_table,
    check_time_interval_time_columns,
    check_time_intervals_stop_after_start,
    check_dynamic_table_region_data_validity,
    check_column_binary_capability,
    check_single_row,
    check_table_values_for_dict,
    check_col_not_nan,
)
from nwbinspector.utils import get_package_version


class TestCheckDynamicTableRegion(TestCase):
    def setUp(self):
        self.table = DynamicTable(name="test_table", description="")
        self.table.add_column(name="test_column", description="")
        for _ in range(10):
            self.table.add_row(test_column=1)

    def test_check_dynamic_table_region_data_validity_lt_zero(self):
        dynamic_table_region = DynamicTableRegion(name="dyn_tab", description="desc", data=[-1, 0], table=self.table)
        assert check_dynamic_table_region_data_validity(dynamic_table_region) == InspectorMessage(
            message="Some elements of dyn_tab are out of range because they are less than 0.",
            importance=Importance.CRITICAL,
            check_function_name="check_dynamic_table_region_data_validity",
            object_type="DynamicTableRegion",
            object_name="dyn_tab",
            location="/",
        )

    def test_check_dynamic_table_region_data_validity_gt_len(self):
        dynamic_table_region = DynamicTableRegion(name="dyn_tab", description="desc", data=[0, 20], table=self.table)
        assert check_dynamic_table_region_data_validity(dynamic_table_region) == InspectorMessage(
            message=(
                "Some elements of dyn_tab are out of range because they are greater than the length of the target "
                "table. Note that data should contain indices, not ids."
            ),
            importance=Importance.CRITICAL,
            check_function_name="check_dynamic_table_region_data_validity",
            object_type="DynamicTableRegion",
            object_name="dyn_tab",
            location="/",
        )

    def test_pass_check_dynamic_table_region_data(self):
        dynamic_table_region = DynamicTableRegion(name="dyn_tab", description="desc", data=[0, 1, 2], table=self.table)
        assert check_dynamic_table_region_data_validity(dynamic_table_region) is None


def test_check_empty_table_with_data():
    table = DynamicTable(name="test_table", description="")
    table.add_column(name="test_column", description="")
    table.add_row(test_column=1)
    assert check_empty_table(table=table) is None


def test_check_empty_table_without_data():
    assert check_empty_table(table=DynamicTable(name="test_table", description="")) == InspectorMessage(
        message="This table has no data added to it.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_empty_table",
        object_type="DynamicTable",
        object_name="test_table",
        location="/",
    )


def test_check_time_interval_time_columns():
    time_intervals = TimeIntervals(name="test_table", description="desc")
    time_intervals.add_row(start_time=2.0, stop_time=3.0)
    time_intervals.add_row(start_time=1.0, stop_time=2.0)

    assert check_time_interval_time_columns(time_intervals) == InspectorMessage(
        message=(
            "['start_time', 'stop_time'] are time columns but the values are not in ascending order. "
            "All times should be in seconds with respect to the session start time."
        ),
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_time_interval_time_columns",
        object_type="TimeIntervals",
        object_name="test_table",
        location="/",
    )


def test_pass_check_time_interval_time_columns():
    time_intervals = TimeIntervals(name="test_table", description="desc")
    time_intervals.add_row(start_time=1.0, stop_time=2.0)
    time_intervals.add_row(start_time=2.0, stop_time=3.0)

    assert check_time_interval_time_columns(time_intervals) is None


def test_check_time_intervals_stop_after_start():
    time_intervals = TimeIntervals(name="test_table", description="desc")
    time_intervals.add_row(start_time=2.0, stop_time=1.5)
    time_intervals.add_row(start_time=3.0, stop_time=1.5)
    assert check_time_intervals_stop_after_start(time_intervals) == InspectorMessage(
        message=(
            "stop_times should be greater than start_times. Make sure the stop times are with respect to the "
            "session start time."
        ),
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_time_intervals_stop_after_start",
        object_type="TimeIntervals",
        object_name="test_table",
        location="/",
    )


def test_pass_check_time_intervals_stop_after_start():
    time_intervals = TimeIntervals(name="test_table", description="desc")
    time_intervals.add_row(start_time=2.0, stop_time=2.5)
    time_intervals.add_row(start_time=3.0, stop_time=3.5)
    assert check_time_intervals_stop_after_start(time_intervals) is None


class TestCheckBinaryColumns(TestCase):
    def setUp(self):
        self.table = DynamicTable(name="test_table", description="")

    def test_non_binary_pass(self):
        self.table.add_column(name="test_col", description="")
        for x in [1.0, 2.0, 3.0]:
            self.table.add_row(test_col=x)
        assert check_column_binary_capability(table=self.table) is None

    def test_array_of_non_binary_pass(self):
        self.table.add_column(name="test_col", description="")
        for x in [[1.0, 2.0], [2.0, 3.0], [1.0, 2.0]]:
            self.table.add_row(test_col=x)
        assert check_column_binary_capability(table=self.table) is None

    def test_jagged_array_of_non_binary_pass(self):
        self.table.add_column(name="test_col", description="", index=True)
        for x in [[1.0, 2.0], [1.0, 2.0, 3.0], [1.0, 2.0]]:
            self.table.add_row(test_col=x)
        assert check_column_binary_capability(table=self.table) is None

    def test_no_saved_bytes_pass(self):
        self.table.add_column(name="test_col", description="")
        for x in np.array([1, 0, 1, 0], dtype="uint8"):
            self.table.add_row(test_col=x)
        assert check_column_binary_capability(table=self.table) is None

    def test_binary_floats_fail(self):
        self.table.add_column(name="test_col", description="")
        for x in [1.0, 0.0, 1.0, 0.0, 1.0]:
            self.table.add_row(test_col=x)
        assert check_column_binary_capability(table=self.table) == [
            InspectorMessage(
                message=(
                    "Column 'test_col' uses 'floats' but has binary values [0. 1.]. Consider making it boolean instead "
                    "and renaming the column to start with 'is_'; doing so will save 35.00B."
                ),
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="check_column_binary_capability",
                object_type="DynamicTable",
                object_name="test_table",
                location="/",
            )
        ]

    def test_binary_int_fail(self):
        self.table.add_column(name="test_col", description="")
        for x in [1, 0, 1, 0, 1]:
            self.table.add_row(test_col=x)
        if platform.system() == "Windows":
            platform_saved_bytes = "15.00B"
        else:
            platform_saved_bytes = "35.00B"
        assert check_column_binary_capability(table=self.table) == [
            InspectorMessage(
                message=(
                    "Column 'test_col' uses 'integers' but has binary values [0 1]. Consider making it boolean "
                    f"instead and renaming the column to start with 'is_'; doing so will save {platform_saved_bytes}."
                ),
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="check_column_binary_capability",
                object_type="DynamicTable",
                object_name="test_table",
                location="/",
            )
        ]

    def test_binary_string_fail(self):
        self.table.add_column(name="test_col", description="")
        for x in ["YES", "NO", "NO", "YES"]:
            self.table.add_row(test_col=x)
        assert check_column_binary_capability(table=self.table) == [
            InspectorMessage(
                message=(
                    "Column 'test_col' uses 'strings' but has binary values ['NO' 'YES']. Consider making it boolean "
                    "instead and renaming the column to start with 'is_'; doing so will save 44.00B."
                ),
                importance=Importance.BEST_PRACTICE_SUGGESTION,
                check_function_name="check_column_binary_capability",
                object_type="DynamicTable",
                object_name="test_table",
                location="/",
            )
        ]

    def test_binary_string_pass(self):
        self.table.add_column(name="test_col", description="")
        for x in ["testing", "testingAgain", "MoreTesting", "testing"]:
            self.table.add_row(test_col=x)
        assert check_column_binary_capability(table=self.table) is None


def test_check_single_row_pass():
    table = DynamicTable(name="test_table", description="")
    table.add_column(name="test_column", description="")
    table.add_row(test_column=1)
    table.add_row(test_column=2)
    assert check_single_row(table=table) is None


def test_check_single_row_ignore_units():
    table = Units(
        name="Units",
    )  # default name when building through nwbfile
    table.add_unit(spike_times=[1, 2, 3])
    assert check_single_row(table=table) is None


def test_check_single_row_ignore_electrodes():
    table = ElectrodeTable(
        name="electrodes",
    )  # default name when building through nwbfile
    if get_package_version(name="pynwb") >= version.Version("2.1.0"):
        table.add_row(
            location="unknown",
            group=ElectrodeGroup(
                name="test_group", description="", device=Device(name="test_device"), location="unknown"
            ),
            group_name="test_group",
        )
    else:
        table.add_row(
            x=np.nan,
            y=np.nan,
            z=np.nan,
            imp=np.nan,
            location="unknown",
            filtering="unknown",
            group=ElectrodeGroup(
                name="test_group", description="", device=Device(name="test_device"), location="unknown"
            ),
            group_name="test_group",
        )
    assert check_single_row(table=table) is None


def test_check_single_row_fail():
    table = DynamicTable(name="test_table", description="")
    table.add_column(name="test_column", description="")
    table.add_row(test_column=1)
    assert check_single_row(table=table) == InspectorMessage(
        message="This table has only a single row; it may be better represented by another data type.",
        importance=Importance.BEST_PRACTICE_SUGGESTION,
        check_function_name="check_single_row",
        object_type="DynamicTable",
        object_name="test_table",
        location="/",
    )


def test_check_table_values_for_dict_non_str():
    table = DynamicTable(name="test_table", description="")
    table.add_column(name="test_column", description="")
    table.add_row(test_column=123)
    assert check_table_values_for_dict(table=table) is None


def test_check_table_values_for_dict_pass():
    table = DynamicTable(name="test_table", description="")
    table.add_column(name="test_column", description="")
    table.add_row(test_column="123")
    assert check_table_values_for_dict(table=table) is None


def test_check_table_values_for_dict_fail():
    table = DynamicTable(name="test_table", description="")
    table.add_column(name="test_column", description="")
    table.add_row(test_column=str(dict(a=1)))
    assert check_table_values_for_dict(table=table)[0] == InspectorMessage(
        message=(
            "The column 'test_column' contains a string value that contains a dictionary! Please unpack "
            "dictionaries as additional rows or columns of the table."
        ),
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_table_values_for_dict",
        object_type="DynamicTable",
        object_name="test_table",
        location="/",
    )


def test_check_table_values_for_dict_json_case_fail():
    table = DynamicTable(name="test_table", description="")
    table.add_column(name="test_column", description="")
    table.add_row(test_column=json.dumps(dict(a=1)))
    assert check_table_values_for_dict(table=table) == [
        InspectorMessage(
            message=(
                "The column 'test_column' contains a string value that contains a dictionary! Please unpack "
                "dictionaries as additional rows or columns of the table. This string is also JSON loadable, so call "
                "`json.loads(...)` on the string to unpack."
            ),
            importance=Importance.BEST_PRACTICE_VIOLATION,
            check_function_name="check_table_values_for_dict",
            object_type="DynamicTable",
            object_name="test_table",
            location="/",
        )
    ]


def test_check_col_not_nan_pass():
    table = DynamicTable(name="test_table", description="")
    for name in ["test_column_not_nan", "test_column_string"]:
        table.add_column(name=name, description="")
    table.add_row(test_column_not_nan=1.0, test_column_string="abc")
    assert check_col_not_nan(table=table) is None


def test_check_col_not_nan_fail():
    """Addition of test_integer_type included from issue 241."""
    table = DynamicTable(name="test_table", description="")
    for name in ["test_column_not_nan_1", "test_column_nan_1", "test_integer_type", "test_column_nan_2"]:
        table.add_column(name=name, description="")
    for _ in range(400):
        table.add_row(
            test_column_not_nan_1=1.0, test_column_nan_1=np.nan, test_integer_type=1, test_column_nan_2=np.nan
        )
    assert check_col_not_nan(table=table) == [
        InspectorMessage(
            message="Column 'test_column_nan_1' might have all NaN values. Consider removing it from the table.",
            importance=Importance.BEST_PRACTICE_SUGGESTION,
            check_function_name="check_col_not_nan",
            object_type="DynamicTable",
            object_name="test_table",
            location="/",
            file_path=None,
        ),
        InspectorMessage(
            message="Column 'test_column_nan_2' might have all NaN values. Consider removing it from the table.",
            importance=Importance.BEST_PRACTICE_SUGGESTION,
            check_function_name="check_col_not_nan",
            object_type="DynamicTable",
            object_name="test_table",
            location="/",
            file_path=None,
        ),
    ]


def test_check_col_not_nan_fail_span_all_data():
    """Addition of test_integer_type included from issue 241."""
    table = DynamicTable(name="test_table", description="")
    for name in ["test_column_not_nan_1", "test_column_nan_1", "test_integer_type", "test_column_nan_2"]:
        table.add_column(name=name, description="")
    for _ in range(180):
        table.add_row(
            test_column_not_nan_1=1.0, test_column_nan_1=np.nan, test_integer_type=1, test_column_nan_2=np.nan
        )
    assert check_col_not_nan(table=table) == [
        InspectorMessage(
            message="Column 'test_column_nan_1' has all NaN values. Consider removing it from the table.",
            importance=Importance.BEST_PRACTICE_SUGGESTION,
            check_function_name="check_col_not_nan",
            object_type="DynamicTable",
            object_name="test_table",
            location="/",
            file_path=None,
        ),
        InspectorMessage(
            message="Column 'test_column_nan_2' has all NaN values. Consider removing it from the table.",
            importance=Importance.BEST_PRACTICE_SUGGESTION,
            check_function_name="check_col_not_nan",
            object_type="DynamicTable",
            object_name="test_table",
            location="/",
            file_path=None,
        ),
    ]



"""Checks specific to intracellular electrophysiology neurodata types."""
from pynwb.icephys import IntracellularElectrode

from ..register_checks import register_check, Importance, InspectorMessage


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=IntracellularElectrode)
def check_intracellular_electrode_cell_id_exists(intracellular_electrode: IntracellularElectrode):
    """Check if the IntracellularElectrode contains a cell_id."""
    if hasattr(intracellular_electrode, "cell_id") and intracellular_electrode.cell_id is None:
        return InspectorMessage(message="Please include a unique cell_id associated with this IntracellularElectrode.")



from jsonschema import ValidationError
from unittest import TestCase

from nwbinspector import (
    Importance,
    check_small_dataset_compression,
    check_regular_timestamps,
    check_data_orientation,
    check_timestamps_match_first_dimension,
    available_checks,
)
from nwbinspector.nwbinspector import validate_config, configure_checks, _copy_function, load_config


class TestCheckConfiguration(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.checks = [
            check_small_dataset_compression,
            check_regular_timestamps,
            check_data_orientation,
            check_timestamps_match_first_dimension,
        ]

    def test_safe_check_copy(self):
        initial_importance = available_checks[0].importance
        changed_check = _copy_function(function=available_checks[0])
        if initial_importance is Importance.CRITICAL:
            changed_importance = Importance.BEST_PRACTICE_SUGGESTION
        else:
            changed_importance = Importance.CRITICAL
        changed_check.importance = changed_importance
        assert available_checks[0].importance is initial_importance
        assert changed_check.importance is changed_importance

    def test_configure_checks_change_importance(self):
        config = dict(
            CRITICAL=["check_small_dataset_compression"],
            BEST_PRACTICE_SUGGESTION=["check_regular_timestamps"],
        )
        checks_out = configure_checks(checks=self.checks, config=config)
        assert (
            checks_out[0].__name__ == "check_small_dataset_compression"
            and checks_out[0].importance is Importance.CRITICAL
        )
        assert (
            checks_out[1].__name__ == "check_regular_timestamps"
            and checks_out[1].importance is Importance.BEST_PRACTICE_SUGGESTION
        )

    def test_configure_checks_no_change(self):
        config = dict(CRITICAL=["check_data_orientation"])
        validate_config(config=config)
        checks_out = configure_checks(checks=self.checks, config=config)
        assert checks_out[2].__name__ == "check_data_orientation", checks_out[2].importance is Importance.CRITICAL

    def test_configure_checks_skip(self):
        config = dict(SKIP=["check_timestamps_match_first_dimension"])
        validate_config(config=config)
        checks_out = configure_checks(checks=self.checks, config=config)
        self.assertListEqual(list1=[x.__name__ for x in checks_out], list2=[x.__name__ for x in self.checks[:3]])

    def test_bad_schema(self):
        config = dict(WRONG="test")
        with self.assertRaises(expected_exception=ValidationError):
            validate_config(config=config)

    def test_load_config(self):
        config = load_config(filepath_or_keyword="dandi")
        self.assertDictEqual(
            d1=config,
            d2=dict(
                CRITICAL=[
                    "check_subject_exists",
                    "check_subject_id_exists",
                    "check_subject_sex",
                    "check_subject_species_exists",
                    "check_subject_species_latin_binomial",
                    "check_subject_age",
                    "check_subject_proper_age_range",
                ],
                BEST_PRACTICE_VIOLATION=[
                    "check_data_orientation",
                ],
            ),
        )



"""Check functions specific to ImageSeries."""
import ntpath
from pathlib import Path

from pynwb.image import ImageSeries

from ..register_checks import register_check, Importance, InspectorMessage
from ..tools import get_nwbfile_path_from_internal_object


@register_check(importance=Importance.CRITICAL, neurodata_type=ImageSeries)
def check_image_series_external_file_valid(image_series: ImageSeries):
    """
    Check if the external_file specified by an ImageSeries actually exists.

    Best Practice: :ref:`best_practice_use_external_mode`
    """
    if image_series.external_file is None:
        return
    nwbfile_path = Path(get_nwbfile_path_from_internal_object(obj=image_series))
    for file_path in image_series.external_file:
        file_path = file_path.decode() if isinstance(file_path, bytes) else file_path
        if not Path(file_path).is_absolute() and not (nwbfile_path.parent / file_path).exists():
            yield InspectorMessage(
                message=(
                    f"The external file '{file_path}' does not exist. Please confirm the relative location to the"
                    " NWBFile."
                )
            )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=ImageSeries)
def check_image_series_external_file_relative(image_series: ImageSeries):
    """
    Check if the external_file specified by an ImageSeries, if it exists, is relative.

    Best Practice: :ref:`best_practice_use_external_mode`
    """
    if image_series.external_file is None:
        return
    for file_path in image_series.external_file:
        file_path = file_path.decode() if isinstance(file_path, bytes) else file_path
        if ntpath.isabs(file_path):  # ntpath required for cross-platform detection
            yield InspectorMessage(
                message=(
                    f"The external file '{file_path}' is not a relative path. "
                    "Please adjust the absolute path to be relative to the location of the NWBFile."
                )
            )


@register_check(importance=Importance.BEST_PRACTICE_VIOLATION, neurodata_type=ImageSeries)
def check_image_series_data_size(image_series: ImageSeries, gb_lower_bound: float = 20.0):
    """
    Check if an ImageSeries stored is larger than gb_lower_bound and suggests external file.

    Best Practice: :ref:`best_practice_use_external_mode`
    """
    data = image_series.data
    data_size_gb = data.size * data.dtype.itemsize / 1e9
    if data_size_gb > gb_lower_bound:
        return InspectorMessage(message=f"ImageSeries {image_series.name} is too large. Use external mode for storage")



from setuptools import setup, find_packages
from pathlib import Path
from shutil import copy

root = Path(__file__).parent
with open(root / "README.md", "r") as f:
    long_description = f.read()
with open(root / "requirements.txt") as f:
    install_requires = f.readlines()
with open(root / "src" / "nwbinspector" / "version.py") as f:
    version = f.read()

# Instantiate the testing configuration file from the base file `base_test_config.json`
base_test_config = Path.cwd() / "base_test_config.json"
local_test_config = Path.cwd() / "tests" / "testing_config.json"
if not local_test_config.exists():
    copy(src=base_test_config, dst=local_test_config)

setup(
    name="nwbinspector",
    version=version.split('"')[1],
    description="Tool to inspect NWB files for best practices compliance.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Ryan Ly, Ben Dichter, and Cody Baker.",
    author_email="rly@lbl.gov, ben.dichter@gmail.com, cody.baker@catalystneuro.com",
    url="https://nwbinspector.readthedocs.io/",
    keywords="nwb",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,  # Includes files described in MANIFEST.in in the installation
    install_requires=install_requires,
    extras_require=dict(dandi=["dandi>=0.39.2"]),
    entry_points={"console_scripts": ["nwbinspector=nwbinspector.nwbinspector:inspect_all_cli"]},
    license="BSD-3-Clause",
)



import pytest

import numpy as np
from pynwb.image import GrayscaleImage

from nwbinspector import InspectorMessage, Importance
from nwbinspector.checks.images import check_order_of_images_unique, check_order_of_images_len

try:
    from pynwb.base import Images, ImageReferences

    HAVE_IMAGES = True
except ImportError:
    HAVE_IMAGES = False
skip_reason = "You must have PyNWB>=v2.1.0 to run these tests!"


@pytest.mark.skipif(not HAVE_IMAGES, reason=skip_reason)
def test_check_order_of_images_unique():

    imgs = [GrayscaleImage(name=f"image{i}", data=np.random.randn(10, 10)) for i in range(5)]
    img_refs = ImageReferences(name="order_of_images", data=imgs + [imgs[0]])
    images = Images(name="my_images", images=imgs, order_of_images=img_refs)

    assert check_order_of_images_unique(images) == InspectorMessage(
        message="order_of_images should have unique values.",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_order_of_images_unique",
        object_type="Images",
        object_name="my_images",
        location="/",
    )


@pytest.mark.skipif(not HAVE_IMAGES, reason=skip_reason)
def test_pass_check_order_of_images_unique():

    imgs = [GrayscaleImage(name=f"image{i}", data=np.random.randn(10, 10)) for i in range(5)]
    img_refs = ImageReferences(name="order_of_images", data=imgs)
    images = Images(name="my_images", images=imgs, order_of_images=img_refs)

    assert check_order_of_images_unique(images) is None


@pytest.mark.skipif(not HAVE_IMAGES, reason=skip_reason)
def test_check_order_of_images_len():

    imgs = [GrayscaleImage(name=f"image{i}", data=np.random.randn(10, 10)) for i in range(5)]
    img_refs = ImageReferences(name="order_of_images", data=imgs + [imgs[0]])
    images = Images(name="my_images", images=imgs, order_of_images=img_refs)

    assert check_order_of_images_len(images) == InspectorMessage(
        message=f"Length of order_of_images (6) does not match the number of images (5).",
        importance=Importance.BEST_PRACTICE_VIOLATION,
        check_function_name="check_order_of_images_len",
        object_type="Images",
        object_name="my_images",
        location="/",
    )


@pytest.mark.skipif(not HAVE_IMAGES, reason=skip_reason)
def test_pass_check_order_of_images_len():

    imgs = [GrayscaleImage(name=f"image{i}", data=np.random.randn(10, 10)) for i in range(5)]
    img_refs = ImageReferences(name="order_of_images", data=imgs)
    images = Images(name="my_images", images=imgs, order_of_images=img_refs)

    assert check_order_of_images_len(images) is None



"""Commonly reused logic for evaluating conditions; must not have external dependencies."""
import os
import re
import json
from typing import TypeVar, Union, Optional, List, Dict, Callable, Tuple
from pathlib import Path
from importlib import import_module
from packaging import version
from time import sleep
from functools import lru_cache

import h5py
import numpy as np
from numpy.typing import ArrayLike


PathType = TypeVar("PathType", str, Path)  # For types that can be either files or folders
FilePathType = TypeVar("FilePathType", str, Path)
OptionalListOfStrings = Optional[List[str]]

dict_regex = r"({.+:.+})"
MAX_CACHE_ITEMS = 1000  # lru_cache default is 128 calls of matching input/output, but might need more to get use here


@lru_cache(maxsize=MAX_CACHE_ITEMS)
def _cache_data_retrieval_command(
    data: h5py.Dataset, reduced_selection: Tuple[Tuple[Optional[int], Optional[int], Optional[int]]]
) -> np.ndarray:
    """LRU caching for _cache_data_selection cannot be applied to list inputs; this expects the tuple or Dataset."""
    selection = tuple([slice(*reduced_slice) for reduced_slice in reduced_selection])  # reconstitute the slices
    return data[selection]


def _cache_data_selection(data: Union[h5py.Dataset, ArrayLike], selection: Union[slice, Tuple[slice]]) -> np.ndarray:
    """Extract the selection lazily from the data object for efficient caching (most beneficial during streaming)."""
    if isinstance(data, np.memmap):  # Technically np.memmap should be able to support this type of behavior as well
        return data[selection]  # But they aren't natively hashable either...
    if not isinstance(data, h5py.Dataset):  # No need to attempt to cache if already an in-memory object
        return np.array(data)[selection]

    # slices also aren't hashable, but their reduced representation is
    if isinstance(selection, slice):  # If a single slice
        reduced_selection = tuple([selection.__reduce__()[1]])  # if a single slice
    else:
        reduced_selection = tuple([selection_slice.__reduce__()[1] for selection_slice in selection])
    return _cache_data_retrieval_command(data=data, reduced_selection=reduced_selection)


def format_byte_size(byte_size: int, units: str = "SI"):
    """
    Format a number representing a total number of bytes into a convenient unit.

    Parameters
    ----------
    byte_size : int
        Total number of bytes to format.
    units : str, optional
        Convention for orders of magnitude to apply.
        May be either SI (orders of 1000) or binary (in memory, orders of 1024).
        The default is SI.
    """
    num = byte_size
    prefixes = ["", "K", "M", "G", "T", "P", "E", "Z"]
    if units == "SI":
        order = 1000.0
        suffix = "B"
    elif units == "binary":
        order = 1024.0
        suffix = "iB"
    else:
        raise ValueError("'units' argument must be either 'SI' (for orders of 1000) or 'binary' (for orders of 1024).")
    for prefix in prefixes:
        if abs(num) < order:
            return f"{num:3.2f}{prefix}{suffix}"
        num /= order
    return f"{num:.2f}Y{suffix}"


def is_regular_series(series: np.ndarray, tolerance_decimals: int = 9):
    """General purpose function for checking if the difference between all consecutive points in a series are equal."""
    uniq_diff_ts = np.unique(np.diff(series).round(decimals=tolerance_decimals))
    return len(uniq_diff_ts) == 1


def is_ascending_series(series: Union[h5py.Dataset, ArrayLike], nelems=None):
    """General purpose function for determining if a series is monotonic increasing."""
    if isinstance(series, h5py.Dataset):
        return np.all(np.diff(_cache_data_selection(data=series, selection=slice(nelems))) > 0)
    else:
        return np.all(np.diff(series[:nelems]) > 0)  # already in memory, no need to cache


def is_dict_in_string(string: str):
    """
    Determine if the string value contains an encoded Python dictionary.

    Can also be the direct results of string casting a dictionary, *e.g.*, ``str(dict(a=1))``.
    """
    return any(re.findall(pattern=dict_regex, string=string))


def is_string_json_loadable(string: str):
    """
    Determine if the serialized dictionary is a JSON object.

    Rather than constructing a complicated regex pattern, a simple try/except of the json.load should suffice.
    """
    try:
        json.loads(string)
        return True
    except json.JSONDecodeError:
        return False


def is_module_installed(module_name: str):
    """
    Check if the given module is installed on the system.

    Used for lazy imports.
    """
    try:
        import_module(name=module_name)
        return True
    except ModuleNotFoundError:
        return False


def get_package_version(name: str) -> version.Version:
    """
    Retrieve the version of a package regardless of if it has a __version__ attribute set.

    Parameters
    ----------
    name : str
        Name of package.

    Returns
    -------
    version : Version
        The package version as an object from packaging.version.Version, which allows comparison to other versions.
    """
    try:
        from importlib.metadata import version as importlib_version

        package_version = importlib_version(name)
    except ModuleNotFoundError:  # Remove the except clause when minimal supported version becomes 3.8
        from pkg_resources import get_distribution

        package_version = get_distribution(name).version
    return version.parse(package_version)


def robust_s3_read(
    command: Callable, max_retries: int = 10, command_args: Optional[list] = None, command_kwargs: Optional[Dict] = None
):
    """Attempt the command (usually acting on an S3 IO) up to the number of max_retries using exponential backoff."""
    command_args = command_args or []
    command_kwargs = command_kwargs or dict()
    for retry in range(max_retries):
        try:
            return command(*command_args, **command_kwargs)
        except Exception as exc:
            if "curl" in str(exc):  # 'cannot curl request' can show up in potentially many different return error types
                sleep(0.1 * 2**retry)
            else:
                raise exc
    raise TimeoutError(f"Unable to complete the command ({command.__name__}) after {max_retries} attempts!")


def calculate_number_of_cpu(requested_cpu: int = 1) -> int:
    """
    Calculate the number CPUs to use with respect to negative slicing and check against maximal available resources.

    Parameters
    ----------
    requested_cpu : int, optional
        The desired number of CPUs to use.

        The default is 1.
    """
    total_cpu = os.cpu_count()
    assert requested_cpu <= total_cpu, f"Requested more CPUs ({requested_cpu}) than are available ({total_cpu})!"
    assert requested_cpu >= -(
        total_cpu - 1
    ), f"Requested fewer CPUs ({requested_cpu}) than are available ({total_cpu})!"
    if requested_cpu > 0:
        return requested_cpu
    else:
        return total_cpu + requested_cpu
